import aiohttp
from herre.grants.oauth2.base import BaseOauth2Grant
from herre.grants.oauth2.session import OAuth2Session
from herre.types import Token
import aiohttp
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from herre.grants.oauth2.session import OAuth2Session
from herre.grants.oauth2.utils import build_token_url
from herre.types import Token


class ClientCredentialsGrant(BaseOauth2Grant):
    async def afetch_token(self, force_refresh: bool = False) -> Token:
        auth_client = BackendApplicationClient(
            client_id=self.client_id.get_secret_value(),
            scope=self.scope_delimiter.join(self.scopes),
        )

        async with OAuth2Session(
            client=auth_client,
            scope=self.scope_delimiter.join(self.scopes),
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:

            token_dict = await session.fetch_token(
                token_url=build_token_url(self),
                client_id=str(self.client_id.get_secret_value()),
                client_secret=str(self.client_secret.get_secret_value()),
                verify=True,
            )

        return Token(**token_dict)



from herre.grants.errors import GrantException


class Oauth2RedirectError(GrantException):
    """Raised when a grant fails to redirect"""

    pass


class Oauth2TimeoutError(GrantException):
    """Raised when a grant times out"""

    pass



import aiohttp
from oauthlib.oauth2 import WebApplicationClient
from herre.grants.oauth2.session import OAuth2Session
import logging
from herre.grants.oauth2.utils import (
    build_authorize_url,
    build_token_url,
    wait_for_redirect,
)
from .base import BaseOauth2Grant
from typing import Awaitable
from herre.types import Token
from typing import Protocol, runtime_checkable

logger = logging.getLogger(__name__)
REDIRECT_PORT = 6767


@runtime_checkable
class RedirectWaiter(Protocol):
    def __call__(
        self,
        starturl,
        redirect_host="localhost",
        redirect_port=6767,
        path="/",
        timeout=400,
    ) -> Awaitable[str]:
        ...


class AuthorizationCodeServerGrant(BaseOauth2Grant):
    redirect_port: int = 6767
    redirect_timeout: int = 40
    redirect_host: str = "localhost"
    redirect_waiter: RedirectWaiter = wait_for_redirect
    """ A simple webserver that will listen for a redirect from the OSF and return the path """

    async def afetch_token(self, force_refresh=True) -> Token:
        web_app_client = WebApplicationClient(
            self.client_id.get_secret_value(),
            scope=self.scope_delimiter.join(self.scopes + ["openid"]),
        )

        # Create an OAuth2 session for the OSF
        async with OAuth2Session(
            self.client_id.get_secret_value(),
            web_app_client,
            scope=self.scope_delimiter.join(self.scopes + ["openid"]),
            redirect_uri=f"http://{self.redirect_host}:{self.redirect_port}/",
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            auth_url, state = session.authorization_url(build_authorize_url(self))

            path = await self.redirect_waiter(
                auth_url,
                redirect_host=self.redirect_host,
                redirect_port=self.redirect_port,
            )

            if path:
                token_dict = await session.fetch_token(
                    build_token_url(self),
                    client_secret=self.client_secret.get_secret_value(),
                    authorization_response=path,
                    state=state,
                )

                return Token(**token_dict)

        raise Exception("Could not fetch token")



from herre.grants.base import BaseGrant
from pydantic import SecretStr, Field
from typing import List
import ssl
import certifi
from ssl import SSLContext


class BaseOauth2Grant(BaseGrant):
    base_url: str
    client_id: SecretStr = SecretStr("")
    client_secret: SecretStr = SecretStr("")
    scopes: List[str] = Field(default_factory=lambda: list(["introspection"]))
    authorize_path: str = "authorize"
    refresh_path: str = "token"
    token_path: str = "token"
    scope_delimiter: str = " "
    allow_insecure: bool = False
    append_trailing_slash: bool = True

    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )



""" Basic types for the herre library

This module contains the basic types for the herre library.

"""

from enum import Enum
from typing import List, Optional
from pydantic import BaseModel
import time

class GrantType(str, Enum):
    """The grant type"""
    CLIENT_CREDENTIALS = "client-credentials"
    AUTHORIZATION_CODE = "authorization-code"


class Token(BaseModel):
    """ A token object"""
    access_token: str
    scope: Optional[List[str]]
    refresh_token: Optional[str]
    expires_in: Optional[int]
    expires_at: Optional[int]
    token_type: Optional[str]

    def is_expired(self) -> bool:
        """Checks if the token is expired"""
        if self.expires_at:
            return self.expires_at < int(time.time())
        return False



from .base import BaseOauth2Grant
from herre.grants.base import BaseGrant, BaseGrantProtocol
import aiohttp
import ssl
from typing import Optional
from herre.types import Token
from .utils import build_refresh_url
import logging

logger = logging.getLogger(__name__)


async def arefresh(
    resfresh_url: str,
    client_id: str,
    client_secret: str,
    refresh_token: str,
    ssl_context: Optional[ssl.SSLContext] = None,
) -> Token:
    """Refreshes a token on the given url with the given client_id and client_secret

    Args:
        resfresh_url (str): The url to refresh the token on
        client_id (str): The client_id to use
        client_secret (str): The client_secret to use
        refresh_token (str): The refresh_token to use
        ssl_context (Optional[ssl.SSLContext], optional): Specific SSL token to use. Defaults to None.

    Returns:
        Oauth2Token: The refreshed token
    """
    assert refresh_token, "Token had not refresh-token attached"
    async with aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(ssl=ssl_context) if ssl_context else None
    ) as session:
        async with session.post(
            resfresh_url,
            data={
                "grant_type": "refresh_token",
                "refresh_token": refresh_token,
                "client_id": client_id,
                "client_secret": client_secret,
            },
        ) as resp:
            token = await resp.json()
            return Token(**token)


class RefreshGrant(BaseGrant):
    """Tries to refresh the token (if it is not expired)
    and a refresh token is available. When the token is expired
    and no refresh token is available, it will try to fetch a new token.

    This grant does not refresh the token automatically. Only when it is
    implicitly called by the Herre api.

    You can choose autofresh grant to refresh the token automatically.

    Args:
        BaseGrant (_type_): _description_
    """

    grant: BaseGrantProtocol

    _token: Optional[Token] = None

    async def afetch_token(self, force_refresh: bool = False) -> Token:
        """Fetches a token from the oauth2 provider.

        Args:
            force_refresh (bool, optional): Force a refresh of the token. Defaults to False.

        Returns:
            Token: The token
        """
        if not force_refresh:
            if self._token and not self._token.is_expired():
                return self._token

            if self._token and self._token.refresh_token:
                try:
                    assert (
                        self._token.refresh_token
                    ), "Token had not refresh-token attached"
                    self._token = await arefresh(
                        resfresh_url=build_refresh_url(self.grant),
                        client_id=self.grant.client_id.get_secret_value(),
                        client_secret=self.grant.client_secret.get_secret_value(),
                        refresh_token=self._token.refresh_token,
                        ssl_context=self.grant.ssl_context,
                    )
                    return self._token
                except Exception:
                    logger.debug("Could not refresh token. Fetching new one")
                    pass

        self._token = await self.grant.afetch_token(force_refresh=force_refresh)
        return self._token



from herre.grants.base import BaseGrant
from herre.types import Token


class StaticGrant(BaseGrant):
    token: Token

    async def afetch_token(self, force_refresh: bool=False) -> Token:
        return self.token



from herre.errors import HerreError


class GrantException(HerreError):
    """Base class for all grant exceptions"""

    pass


class RetriesExceededException(GrantException):
    """Raised when a grant exceeds the number of retries"""

    pass



from typing import Dict, Type, Optional

from pydantic import BaseModel
from herre.grants.base import BaseGrant
from herre.types import GrantType
import logging


logger = logging.getLogger(__name__)


class GrantRegistry(BaseModel):
    _registered_grants: Optional[Dict[GrantType, Type[BaseGrant]]] = None

    def register_grant(self, type: GrantType, grant: Type[BaseGrant]) -> None:
        if not self._registered_grants:
            self._registered_grants = {}
        assert hasattr(
            grant, "afetch_token"
        ), f"Grant {grant}must implement afetchtoken"
        self._registered_grants[type] = grant

    def get_grant_for_type(self, type) -> Type[BaseGrant]:
        return self._registered_grants[type]

    class Config:
        json_encoders = {type: lambda x: x.__name__}
        underscore_attrs_are_private = True


def register_grant(type: GrantType):
    def real_decorator(grant):
        get_default_grant_registry().register_grant(type, grant)
        return grant

    return real_decorator


GRANT_REGISTRY = None


def get_default_grant_registry():
    global GRANT_REGISTRY
    if not GRANT_REGISTRY:
        GRANT_REGISTRY = GrantRegistry()
        from herre.grants.oauth2.client_credentials import ClientCredentialsGrant
        from herre.grants.oauth2.authorization_code_server import (
            AuthorizationCodeServerGrant,
        )

        GRANT_REGISTRY.register_grant(
            GrantType.AUTHORIZATION_CODE, AuthorizationCodeServerGrant
        )
        GRANT_REGISTRY.register_grant(
            GrantType.CLIENT_CREDENTIALS, ClientCredentialsGrant
        )

    return GRANT_REGISTRY



from abc import abstractmethod

from pydantic import BaseModel
from herre.types import Token
from typing import Protocol, runtime_checkable
import logging

logger = logging.getLogger(__name__)


@runtime_checkable
class BaseGrantProtocol(Protocol):

    async def afetch_token(self, force_refresh: bool =False) -> Token:
        ...


class BaseGrant(BaseModel):

    @abstractmethod
    async def afetch_token(self, force_refresh: bool =False) -> Token:
        raise NotImplementedError("Implement afetch_token")

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        extras = "forbid"



from herre.grants.qt.login_screen import (
    QtLoginScreen,
    Token,
    User,
    MalformedAnswerException,
    FetchingUserException,
)
from typing import Optional
from fakts import Fakts
import logging
import aiohttp
from pydantic import Field

logger = logging.getLogger(__name__)


class FaktsQtLoginScreen(QtLoginScreen):
    fakts: Fakts
    userinfo_endpoint: Optional[str]
    fakts_group: str = "lok"
    fakts: Optional[Fakts] = None
    fetch_timeout: int = Field(default=3, description="Timeout for fetching the user")

    _userinfo_endpoint = None

    """ An ssl context to use for the connection to the endpoint"""

    async def aget_userinfo_endpoint(self) -> str:
        unserialized = await self.fakts.aget(self.fakts_group)
        self.userinfo_endpoint = unserialized["base_url"] + "/me/"
        return unserialized["base_url"] + "/me/"

    async def afetch_user(self, token: Token) -> User:
        if not self._userinfo_endpoint:
            self._userinfo_endpoint = await self.aget_userinfo_endpoint()

        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
            headers={"Authorization": f"Bearer {token.access_token}"},
        ) as session:
            async with session.get(
                f"{self.userinfo_endpoint}",
                timeout=self.fetch_timeout,
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if "username" not in data:
                        logger.error(f"Malformed answer: {data}")
                        raise MalformedAnswerException("Malformed Answer")

                    return User(**data)

                else:
                    raise FetchingUserException(
                        "Error! Coud not retrieve on the endpoint"
                    )



import ssl
import certifi

from herre.grants.base import BaseGrant, BaseGrantProtocol
from herre.types import Token
from typing import List, Optional, Dict
import logging
from qtpy import QtWidgets, QtCore
from pydantic import BaseModel, Field
logger = logging.getLogger(__name__)
from koil.qt import QtCoro, QtFuture
from .errors import UserCancelledError
import asyncio
import aiohttp
from herre.grants.errors import GrantException
import json

class User(BaseModel):
    username: str
    email: str
    first_name: Optional[str]
    last_name: Optional[str]
    avatar: Optional[str]
    sub: str
    id: str


class UserStore(BaseModel):
    user: User
    token: Token

class Storage(BaseModel):
    users: Dict[str,UserStore]


class UserWidget(QtWidgets.QWidget):
    login_clicked = QtCore.Signal(UserStore)
    logout_clicked = QtCore.Signal(UserStore)

    def __init__(self, store: UserStore, *args,  **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.store = store

        self.user_label = QtWidgets.QLabel(store.user.username)
        self.user_label.mousePressEvent = self.on_clicked

        self.logout_button = QtWidgets.QPushButton("Logout")
        self.logout_button.clicked.connect(self.on_logout_clicked)

        self.hlayout = QtWidgets.QHBoxLayout()
        self.setLayout(self.hlayout)
        self.hlayout.addWidget(self.user_label)
        self.hlayout.addWidget(self.logout_button)

    def on_clicked(self, event):
        self.login_clicked.emit(self.store)

    def on_logout_clicked(self):
        self.logout_clicked.emit(self.store)



class LoginWidget(QtWidgets.QDialog):


    def __init__(self, identifier, version, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.settings = QtCore.QSettings("Arkitekt", f"{identifier}:{version}")
        
        self.vlayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.vlayout)


        self.user_list_l = QtWidgets.QVBoxLayout()
        self.vlayout.addLayout(self.user_list_l)

        self.show_coro = QtCoro(self.show, autoresolve=True)
        self.hide_coro = QtCoro(self.hide, autoresolve=True)

        self.select_user_coro = QtCoro(self.start_select_user)
        self.store_user_coro = QtCoro(self.store_user, autoresolve=True)

        self.button_layout = QtWidgets.QHBoxLayout()

        self.new_user_button = QtWidgets.QPushButton("New User")
        self.button_layout.addWidget(self.new_user_button)
        self.new_user_button.clicked.connect(self.on_new_user_clicked)

        self.cancel_button = QtWidgets.QPushButton("Cancel")
        self.button_layout.addWidget(self.cancel_button)
        self.cancel_button.clicked.connect(self.on_cancel_clicked)


        self.vlayout.addLayout(self.button_layout)

        self.refresh_users()
        self._future: Optional[QtFuture] = None

    def start_select_user(self, future: QtFuture):
        self._future = future

    def store_user(self, userStore: UserStore) -> None:
        un_storage = self.settings.value("userstore", None)
        if un_storage:
            storage = Storage(**json.loads(un_storage))
        else:
            storage = Storage(users={})
        storage.users[userStore.user.id] = userStore

        self.settings.setValue("userstore", storage.json())
        self.refresh_users()

    def delete_user(self, userStore: UserStore) -> None:
        un_storage = self.settings.value("userstore", None)
        if un_storage:
            storage = Storage(**json.loads(un_storage))
        else:
            storage = Storage(users={})
        del storage.users[userStore.user.id]
        self.settings.setValue("userstore", storage.json())
        self.refresh_users()

    def get_user_stores(self) -> List[UserStore]:
        un_storage = self.settings.value("userstore", None)
        if not un_storage:
            return []
        storage = Storage(**json.loads(un_storage))
        return [userstore for userstore in storage.users.values()]

    def clearLayout(self, layout) -> None:
        while layout.count():
            child = layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def refresh_users(self):
        self.clearLayout(self.user_list_l)
        self.user_list_l.addWidget(QtWidgets.QLabel("Select User"))
        for userstore in self.get_user_stores():
            widget = UserWidget(userstore)
            self.user_list_l.addWidget(widget)
            widget.login_clicked.connect(self.on_user_clicked)
            widget.logout_clicked.connect(self.delete_user)

    def on_cancel_clicked(self):
        self.close()

    def close(self) -> bool:
        if self._future:
            self._future.reject(UserCancelledError("The user cancelled the login"))
        return super().close()

    def on_new_user_clicked(self) -> None:
        if self._future:
            self._future.resolve()

    def on_user_clicked(self, user: UserStore) -> None:
        if self._future:
            self._future.resolve(user)


    def retrieve_users(self) -> List[User]:
        return self.settings.value("user_list", [])


class FetchingUserException(GrantException):
    """A base exception for errors that occur while fetching a user.

    Args:
        GrantException (_type_): _description_
    """
    pass

class MalformedAnswerException(FetchingUserException):
    """Raised when the answer from the userinfo endpoint is malformed."""
    pass



class QtLoginScreen(BaseGrant):
    """A grant that uses a Qt login screen to authenticate the user.

    The user is presented with a login screen that allows them to select a user
    from a list of previously logged in users. If the user is not in the list,
    they can click a button to start the login flow.


    """

    grant: BaseGrantProtocol
    """The grant to use for the login flow."""
    userinfo_endpoint: str
    """The endpoint to use for fetching the user info."""
    widget: LoginWidget
    """The widget to use for the login screen."""
    auto_login: bool = False
    """If true, the user will be automatically logged in if there is only one user in the list."""
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    async def afetch_user(self, token: Token) -> User:
       async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
            headers={"Authorization": f"Bearer {token.access_token}"},
        ) as session:
            async with session.get(
                f"{self.userinfo_endpoint}",
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if "username" not in data:
                        logger.error(f"Malformed answer: {data}")
                        raise MalformedAnswerException("Malformed Answer")

                    return User(**data)

                else:
                    raise FetchingUserException("Error! Coud not retrieve on the endpoint")



    async def afetch_token(self, force_refresh: bool =False) -> Token:
        """Fetches the token

        This function will only delegate to the grant if the user has not
        previously logged in (aka there is no token in the storage) Or if the
        force_refresh flag is set.

        Args:
            force_refresh (bool, optional): _description_. Defaults to False.

        Raises:
            e: _description_

        Returns:
            Token: _description_
        """

        try:
            await self.widget.show_coro.acall()

            store: UserStore = await self.widget.select_user_coro.acall()
            if store and not force_refresh:
                await self.widget.hide_coro.acall()
                return store.token

            new_token = await self.grant.afetch_token(force_refresh=force_refresh)
            user = await self.afetch_user(new_token) 

            await self.widget.store_user_coro.acall(UserStore(user=user, token=new_token))


            await self.widget.hide_coro.acall()
            return new_token

        except asyncio.CancelledError as e:
            await self.widget.hide_coro.acall()
            raise e

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True






from typing import Optional
import aiohttp
from pydantic import Field
from herre.grants.oauth2.base import BaseOauth2Grant
from herre.grants.oauth2.session import OAuth2Session
from herre.types import Token
from qtpy import QtWidgets
from koil.qt import QtCoro, QtFuture
from qtpy.QtWebEngineWidgets import QWebEngineView
from qtpy import QtCore
from herre.grants.oauth2.utils import build_authorize_url, build_token_url

from oauthlib.oauth2 import WebApplicationClient


class LoginWrapper(QWebEngineView):
    redirect_uri: Optional[str] = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.name = self.page()
        self.redirect_uri = None
        self.show_coro = QtCoro(self.initialize)

    def initialize(self, future: QtFuture, auth_url: str, redirect_uri: str) -> None:
        self.future = future
        self.redirect_uri = redirect_uri
        self.future = future

        self.load(QtCore.QUrl(auth_url))
        self.show()

    def _interceptUrl(self, url: QtCore.QUrl):
        url_string = bytes(url.toEncoded()).decode()
        if self.redirect_uri:
            if url_string.startswith(self.redirect_uri):
                if self.future:
                    self.future.resolve(url_string)
                    self.close()


class AuthorizationCodeQtGrant(BaseOauth2Grant):
    redirect_port: int = 6767
    redirect_host: str = "localhost"
    redirect_timeout: int = 40
    login_wrapper: LoginWrapper = Field(default_factory=LoginWrapper)

    async def afetch_token(self, force_refresh: bool = False) -> Token:
        redirect_uri = f"http://{self.redirect_host}:{self.redirect_port}"

        web_app_client = WebApplicationClient(
            self.client_id.get_secret_value(),
            scope=self.scope_delimiter.join(self.scopes + ["openid"]),
        )

        async with OAuth2Session(
            self.client_id.get_secret_value(),
            web_app_client,
            scope=self.scope_delimiter.join(self.scopes + ["openid"]),
            redirect_uri=redirect_uri,
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            auth_url, state = session.authorization_url(build_authorize_url(self))

            path = await self.login_wrapper.show_coro.acall(auth_url, redirect_uri)

            token_dict = await session.fetch_token(
                build_token_url(self),
                client_secret=self.client_secret.get_secret_value(),
                authorization_response=path,
                state=state,
            )

            return Token(**token_dict)

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        json_encoder = {
            QtWidgets.QWidget: lambda x: x.__class__.__name__,
        }



from .base import BaseOauth2Grant
from aiohttp import web
import asyncio
import webbrowser
import logging
from .errors import Oauth2RedirectError


logger = logging.getLogger(__name__)


def build_authorize_url(grant: BaseOauth2Grant) -> str:
    return (
        f"{grant.base_url}/{grant.authorize_path}/"
        if grant.append_trailing_slash
        else f"{grant.base_url}/{grant.authorize_path}"
    )


def build_token_url(grant: BaseOauth2Grant) -> str:
    return (
        f"{grant.base_url}/{grant.token_path}/"
        if grant.append_trailing_slash
        else f"{grant.base_url}/{grant.token_path}"
    )


def build_refresh_url(grant: BaseOauth2Grant) -> str:
    return (
        f"{grant.base_url}/{grant.refresh_path}/"
        if grant.append_trailing_slash
        else f"{grant.base_url}/{grant.refresh_path}"
    )


success_full_return = """
<!doctype html>

<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">

  <title>Sucessfull Login</title>
  <meta name="description" content="A success page for login">
  <meta name="author" content="SitePoint">

  <meta property="og:title" content="Sucessfull Login">
  <meta property="og:type" content="website">

</head>

<body>
    <h1>Successfull Login</h1>
    <p>You can close this window</p>
  <script type="text/javascript">
  window.close() ;
</script>
</body>
</html>
"""

failure_return = """
<!doctype html>

<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">

  <title>Sucessfull Login</title>
  <meta name="description" content="A success page for login">
  <meta name="author" content="SitePoint">

  <meta property="og:title" content="Sucessfull Login">
  <meta property="og:type" content="website">

</head>

<body>
    <h1>Unsucessfull login</h1>
    <p>You can close this window</p>
  <script type="text/javascript">
  window.close() ;
</script>
</body>
</html>
"""


def wrapped_qs_future(future):
    async def web_token_response(request):
        try:
            future.set_result(request.path_qs)
            return web.Response(text=success_full_return, content_type="text/html")
        except Exception as e:
            return web.Response(text=failure_return, content_type="text/html")

    return web_token_response


async def wait_for_redirect(
    starturl,
    redirect_host="localhost",
    redirect_port=6767,
    path="/",
    timeout=400,
    print_function=False,
    handle_signals=False,
) -> str:
    """A simple webserver that will listen for a redirect from the OSF and return the path"""

    token_future = asyncio.get_event_loop().create_future()

    app = web.Application()

    app.router.add_get(path, wrapped_qs_future(token_future))

    webserver_future = asyncio.wait_for(
        web._run_app(
            app,
            host=redirect_host,
            port=redirect_port,
            print=lambda x: logger.info(x),
            handle_signals=handle_signals,
        ),
        timeout,
    )

    webserver_task = asyncio.create_task(webserver_future)

    webbrowser.open(starturl)
    done, pending = await asyncio.wait(
        [token_future, webserver_task], return_when=asyncio.FIRST_COMPLETED
    )

    for tf in done:
        if tf == token_future:
            redirect_qs = tf.result()
        else:
            raise Oauth2RedirectError(f"Webserver ended unexpectedly {str(tf.exception())}")

    for task in pending:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

    if not redirect_qs:
        raise Oauth2RedirectError("Webserver")

    return redirect_qs



import asyncio
from typing import Optional, TypeVar
from herre.errors import HerreError
from herre.grants.base import BaseGrant
import os
import logging
from herre.types import Token
import contextvars
from koil.composition import KoiledModel
from koil.helpers import unkoil

current_herre: contextvars.ContextVar["Herre"] = contextvars.ContextVar("current_herre")

logger = logging.getLogger(__name__)

T = TypeVar("T")

class Herre(KoiledModel):
    """Herre is a client for Token authentication.

    It provides a unified, composable interface for token based authentication based on grant.
    A grant is a class that is able to retrieve a token. Importantly grants do not have to
    directly call the token endpoint. They can also use a cache or other means to retrieve the
    token.

    Herre is a context manager. This allows it both to provide itself as a singleton and handle
    the asynchronous interface of the grant. As well as providing a lock to ensure that only one
    request is happening at a time.

    Example:
        ```python
        from herre import Herre,
        from herre.grants.oauth2.client_credentials import ClientCredentialsGrant

        herre = Herre(
            grant=ClientCredentialsGrant(
                client_id="my_client_id",
                client_secret="my_client
                base_url="https://my_token_url",
            )
        )

        with herre:
            token = herre.get_token()
        ```

        or aync

        ```python
        from herre import Herre,
        from herre.grants.oauth2.client_credentials import ClientCredentialsGrant

        herre = Herre(
            grant=ClientCredentialsGrant(
                client_id="my_client_id",
                client_secret="my_client
                base_url="https://my_token_url",
            )
        )

        async with herre:
            token = await herre.get_token()
        ```

    """

    grant: BaseGrant
    max_retries: int = 1
    allow_insecure: bool = False
    scope_delimiter: str = " "
    auto_login = True

    login_on_enter: bool = False
    logout_on_exit: bool = False
    entered = False

    no_temp: bool = False

    _lock: Optional[asyncio.Lock] = None
    _token: Optional[Token] = None

    @property
    def token(self) -> Token:
        "The current token"
        assert (
            self._lock is not None
        ), "Please enter the context first to access this variable"
        assert self._token is not None, "No token fetched"
        return self._token

    async def aget_token(self, force_refresh: bool = False) -> str:
        """Get an access token

        Will return an access token if it is already available or
        try to login depending on auto_login. The checking and potential retrieving will happen
        in a lock ensuring that not multiple requests are happening at the same time.

        Args:
            auto_login (bool, optional): Should we allow an automatic login. Defaults to True.

        Returns:
            str:  The access token
        """
        assert (
            self._lock is not None
        ), "We were not initialized. Please enter the context first"

        async with self._lock:
            if not self._token or not self._token.access_token or force_refresh:
                if not self.auto_login:
                    raise HerreError(
                        "Auto-login is set to false and we need to login again"
                    )
                await self.alogin(force_refresh=force_refresh)

        assert self._token is not None, "We should have a token by now"
        return self._token.access_token

    async def arefresh_token(self) -> str:
        """Refresh the token

        Will try to refresh the token. If the token is not refreshable, it will try to login again.
        """
        return await self.aget_token(force_refresh=True)

    async def alogin(self, force_refresh: bool = False) -> Token:
        """Async Login 

        Login the client. This will try to login the client and set the token. If the token is
        not refreshable, it will try to login again.

        Args:
            force_refresh (bool, optional): Should we tell grants to refresh . Defaults to False.

        Raises:
            AssertionError: If we are not initialized

        """
        assert (
            self._lock is not None
        ), "We were not initialized. Please enter the context first"

        potential_token = await self.grant.afetch_token(force_refresh=force_refresh)
        self._token = potential_token
        return self._token

    async def alogout(self) -> None:
        """Logout 

        Logout will set the current state to none and delete the token file.
        """
        assert (
            self._lock is not None
        ), "We were not initialized. Please enter the context first"

        self._token = None

    def login(self, force_refresh: bool = False, retry: int = 0) -> Token:
        """Login

        Login the client. This will try to login the client and set the token. If the token is
        not refreshable, it will try to login again.

        Args:
            force_refresh (bool, optional): Should we tell grants to refresh . Defaults to False.

        Raises:
            AssertionError: If we are not initialized

        """
        return unkoil(
            self.alogin,
            force_refresh=force_refresh,
            retry=retry,
        )

    def get_token(self, force_refresh: bool = False) -> str:
        """Get an access token

        Will return an access token if it is already available or
        try to login depending on auto_login. The checking and potential retrieving will happen
        in a lock ensuring that not multiple requests are happening at the same time.
        """
        return unkoil(self.aget_token, force_refresh=force_refresh)

    def logout(self) -> None:
        """Logout 
        
        Logout will set the current state to none and delete the token file.
        """
        return unkoil(self.alogout)

    @property
    def logged_in(self) -> bool:
        """Is the client logged in
        """
        return self._token is not None

    async def __aenter__(self: T) -> T:
        """ Enters the context and logs in if needed"""
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if self.allow_insecure else "1"
        self._lock = asyncio.Lock()
        current_herre.set(self)
        if self.login_on_enter:
            await self.alogin()
        self.entered = True
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exits the context and logs out if needed"""
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0"
        if self.logout_on_exit:
            await self.alogout()
        current_herre.set(None)

    def _repr_html_inline_(self) -> str:
        """ Jupyter inline representation """
        return f"<table><tr><td>auto_login</td><td>{self.auto_login}</td></tr></table>"

    class Config:
        
        underscore_attrs_are_private = True
        extra = "forbid"



from .client_credentials import ClientCredentialsGrant
from .refresh import RefreshGrant
from .authorization_code_server import AuthorizationCodeServerGrant

__all__ = ["ClientCredentialsGrant", "RefreshGrant", "AuthorizationCodeServerGrant"]



from typing import Any, Dict, List, Optional, Type, runtime_checkable, Protocol
from pydantic import BaseModel, Field, SecretStr
from herre.grants.base import BaseGrant
from herre.grants.registry import GrantRegistry, get_default_grant_registry
from herre.types import GrantType
from herre.grants.oauth2.base import BaseOauth2Grant
from oauthlib.oauth2.rfc6749.errors import InvalidClientError
from fakts import Fakts


class HerreFakt(BaseModel):
    base_url: str
    name: str
    client_id: SecretStr
    client_secret: SecretStr
    grant_type: GrantType
    grant_kwargs: Dict[str, Any] = {}
    scopes: List[str]
    redirect_uri: Optional[str]
    username: Optional[str]
    password: Optional[str]
    timeout: int = 500
    no_temp: bool = False


@runtime_checkable
class BaseGrantFactory(Protocol):
    def __call__(self, **kwds: Any) -> BaseGrant:
        pass


class FaktsGrant(BaseOauth2Grant):
    fakts: Fakts
    base_url: Optional[str] = None
    grant_class: Optional[BaseGrantFactory]
    grant_registry: GrantRegistry = Field(default_factory=get_default_grant_registry)
    fakts_group: str
    allow_reconfiguration_on_invalid_client: bool = True

    _configured = False
    _activegrant: Optional[BaseGrant] = None
    _old_fakt: Dict[str, Any] = {}

    def configure(self, fakt: HerreFakt) -> None:
        grant_class = self.grant_class or self.grant_registry.get_grant_for_type(
            fakt.grant_type
        )

        self._activegrant = grant_class(**fakt.dict())

    async def afetch_token(self, force_refresh=False):
        self.fakts

        if self.fakts.has_changed(self._old_fakt, self.fakts_group):
            self._old_fakt = await self.fakts.aget(self.fakts_group)
            self.configure(HerreFakt(**self._old_fakt))

        try:
            return await self._activegrant.afetch_token(force_refresh=force_refresh)
        except InvalidClientError as e:
            if self.allow_reconfiguration_on_invalid_client:
                self._old_fakt = await self.fakts.aget(
                    self.fakts_group, force_refresh=True
                )
                self.configure(HerreFakt(**self._old_fakt))
                return await self._activegrant.afetch_token(force_refresh=True)
            else:
                raise e

    async def __aenter__(self, **kwargs):
        return await super().__aenter__(**kwargs)



from herre.grants.base import BaseGrant
from herre.types import Token


class MockGrant(BaseGrant):
    async def afetch_token(self, force_refresh=False) -> Token:
        return Token(access_token="mock_token", refresh_token="mock_refresh_token")



class HerreError(Exception):
    """Base class for Herre errors"""

    pass


class NoHerreFound(HerreError):
    """Raised when no Herre instance is found in the context."""

    pass


class LoginException(HerreError):
    """Raised when the login dfails."""

    pass


class ConfigurationException(HerreError):
    """Raised when the configuration is invalid."""

    pass
